use std::sync::Arc;

use axum::{Extension, Json};
use eyre::Context;
use foxchat::{
    fed,
    http::ApiError,
    s2s::http::{HelloRequest, HelloResponse, NodeResponse},
    signature::verify_signature,
    FoxError,
};
use rsa::{
    pkcs1::{DecodeRsaPublicKey, EncodeRsaPublicKey},
    RsaPublicKey,
};
use tracing::error;
use ulid::Ulid;

use crate::{app_state::AppState, fed::FoxSignatureData};

pub async fn post_hello(
    Extension(state): Extension<Arc<AppState>>,
    signature: FoxSignatureData,
    Json(data): Json<HelloRequest>,
) -> Result<Json<HelloResponse>, ApiError> {
    let node = fed::get::<NodeResponse>(
        &state.private_key,
        &state.config.domain,
        &data.host,
        "/_fox/ident/node",
        None,
    )
    .await?;
    let public_key =
        RsaPublicKey::from_pkcs1_pem(&node.public_key).wrap_err("parsing remote public key")?;

    if state.config.domain.clone() != signature.host {
        return Err(FoxError::InvalidHeader.into());
    }

    if let Err(e) = verify_signature(
        &public_key,
        signature.signature,
        signature.date,
        &signature.host,
        &signature.request_path,
        signature.content_length,
        signature.user_id,
    ) {
        error!(
            "Verifying signature from hello request from {}: {}",
            signature.domain, e
        );

        return Err(FoxError::InvalidSignature.into());
    }

    sqlx::query!(
        "insert into identity_instances (id, domain, base_url, public_key) values ($1, $2, $3, $4)",
        Ulid::new().to_string(),
        signature.domain,
        format!("https://{}", signature.domain),
        node.public_key,
    )
    .execute(&state.pool)
    .await?;

    Ok(Json(HelloResponse {
        public_key: state
            .public_key
            .to_pkcs1_pem(rsa::pkcs8::LineEnding::CR)
            .wrap_err("formatting instance public key")?,
        host: state.config.domain.clone(),
    }))
}
