use std::sync::Arc;

use eyre::{Result, Context};
use foxchat::{fed::request::is_valid_domain, FoxError};
use rsa::{RsaPublicKey, pkcs1::DecodeRsaPublicKey};
use serde::{Deserialize, Serialize};

use crate::app_state::AppState;

#[derive(Serialize, Clone, Debug)]
pub struct IdentityInstance {
    pub id: String,
    pub domain: String,
    pub base_url: String,
    pub public_key: String,
    pub status: InstanceStatus,
    pub reason: Option<String>,
}

#[derive(Serialize, Deserialize, Clone, Debug, sqlx::Type)]
#[sqlx(type_name = "instance_status", rename_all = "lowercase")]
pub enum InstanceStatus {
    Active,
    Suspended,
}

impl IdentityInstance {
    pub async fn get(state: Arc<AppState>, domain: &str) -> Result<Self> {
        if !is_valid_domain(domain) {
            return Err(FoxError::InvalidServer.into());
        }

        if let Some(instance) = sqlx::query_as!(
            Self,
            r#"select id, domain, base_url, public_key,
            status as "status: InstanceStatus", reason
            from identity_instances where domain = $1"#,
            domain
        )
        .fetch_optional(&state.pool)
        .await?
        {
            return Ok(instance);
        }

        return Err(FoxError::InvalidServer.into());
    }

    pub fn parse_public_key(&self) -> Result<RsaPublicKey> {
        RsaPublicKey::from_pkcs1_pem(&self.public_key).wrap_err("parsing identity instance public key")
    }
}
