use std::sync::Arc;

use eyre::Result;
use foxchat::{fed, model::User as HttpUser, Id, id::UserType};
use ulid::Ulid;

use crate::app_state::AppState;

use super::identity_instance::IdentityInstance;

pub struct User {
    pub id: Id<UserType>,
    pub instance_id: String,
    pub instance: IdentityInstance,
    pub remote_user_id: String,
    pub username: String,
    pub avatar: Option<String>,
}

impl User {
    pub async fn get(
        state: &Arc<AppState>,
        instance: &IdentityInstance,
        remote_id: &str,
    ) -> Result<User> {
        if let Some(user) = sqlx::query!(
            "select * from users where instance_id = $1 and remote_user_id = $2",
            instance.id,
            remote_id
        )
        .fetch_optional(&state.pool)
        .await?
        {
            return Ok(User {
                id: user.id.into(),
                username: user.username,
                instance_id: user.instance_id,
                instance: instance.to_owned(),
                remote_user_id: user.remote_user_id,
                avatar: user.avatar,
            });
        }

        let http_user = fed::get::<HttpUser>(
            &state.private_key,
            &state.config.domain,
            &instance.domain,
            &format!("/_fox/ident/users/{}", remote_id),
            None,
        )
        .await?;

        let user = sqlx::query!(
            "insert into users (id, instance_id, remote_user_id, username, avatar) values ($1, $2, $3, $4, $5) returning *",
            Ulid::new().to_string(),
            instance.id,
            http_user.id,
            http_user.username,
            http_user.avatar_url
        )
        .fetch_one(&state.pool)
        .await?;

        Ok(User {
            id: user.id.into(),
            username: user.username,
            instance_id: user.instance_id,
            instance: instance.to_owned(),
            remote_user_id: user.remote_user_id,
            avatar: user.avatar,
        })
    }
}
